/*
 * Copyright (c) 2004 Damir Bikmuhametov <boco@ufanet.ru>
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: avecomm.c,v 1.2 2004/02/09 19:57:37 boco Exp $
 */
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>

#include "common.h"
#include "avecomm.h"

ssize_t timed_read(int, void *, size_t, int);
ssize_t timed_write(int, const void *, size_t, int);
int ave_connect(char *);
void ave_disconnect(int);
void free_array(void **, int);
char * extract_virus_name(char *);
char * getline(char **);

/* main scanning procedure */

int ave_scan_file(char * filename, char ** viruslist)
{
	char * id = "ave_scan_file";
	char buffer[4096];
	int fd = -1;
	int len;
	int infected = 0;
	int end_of_conversation = 0;
	char * p, * q;
	char * virus_names[11];
	char ** virus_name;
	int viruses = 1;

	if (!filename || *filename == '\0') {
		syslog(LOG_ERR, "%s: empty filename", id);
		return -1;
	}

	if (!ave_socket || *ave_socket == '\0') {
		syslog(LOG_ERR, "%s: empty AVE socket name", id);
		return -1;
	}

	/* connect to aveserver */

	if ((fd = ave_connect(ave_socket)) == -1) {
		syslog(LOG_ERR, "%s: error connecting to AVE socket: %m", id);
		return -1;
	}

	memset(buffer, 0, sizeof(buffer));

	if ((len=timed_read(fd,buffer, sizeof(buffer)-1,ave_timeout)) == -1 ) {
		syslog(LOG_ERR, "%s: can't read AVE banner: %m", id);
		ave_disconnect(fd);
		return -1;
	}

	buffer[len + 1] = 0;
	if ((p = strchr(buffer, '\n'))) *p = 0;
	if ((p = strchr(buffer, '\r'))) *p = 0;

	if (memcmp(buffer, "201", 3)) {
		syslog(LOG_ERR, "%s: invalid AVE banner: %s", id, buffer);
		ave_disconnect(fd);
		/* FIXME: process 4xx codes more gently */
		return -1;
	}

	/* prepare array for virus names */

	memset(virus_names, 0, sizeof(virus_names));
	virus_name = virus_names;

	/* prepare AVE command */

	snprintf(buffer, sizeof(buffer) - 1, "SCAN xmQPRSTUWb %s\r\n",
		       filename);

	len = strlen(buffer);
	
	if (timed_write(fd, buffer, len, ave_timeout) != len) {
		syslog(LOG_ERR, "%s: can't write to AVE socket: %m", id);
		ave_disconnect(fd);
		return -1;
	}

	while (!end_of_conversation &&
		(len=timed_read(fd, buffer, sizeof(buffer)-1, ave_timeout))) {

		if (len == -1) {
			syslog(LOG_ERR, "%s: can't read from AVE socket: %m",
				       id);
			free_array((void **) virus_names,
			       sizeof(virus_names) / sizeof(char *));
			ave_disconnect(fd);
			return -1;
		}

		buffer[len] = 0;
		p = buffer;
		while ((q = getline(&p))) {
			if (*q == '5') {
				syslog(LOG_ERR, "%s: error from AVE: %s",
					       id, q);
				infected = -1;
				end_of_conversation++;
				break;
			} else if (*q == '2') {
				*(q + 3) = 0;
				switch (atoi(q)) {
					case 221:
					case 230:
					case 231:
					case 232:	infected = 1;
				}
				end_of_conversation++;
				break;
			} else if (!memcmp(q, "322", 3) &&
				viruses < sizeof(virus_names)/sizeof(char*)) {
				*virus_name = extract_virus_name(q);
				if (*virus_name) {
					virus_name++;
					viruses++;
				}
			}
		}
	}
	
	strcpy(buffer, "QUIT\r\n");

	len = strlen(buffer);

	if (timed_write(fd, buffer, len, ave_timeout) != len) {
		syslog(LOG_ERR, "%s: can't write to AVE socket: %m", id);
		free_array((void **) virus_names,
			       sizeof(virus_names) / sizeof(char *));
		ave_disconnect(fd);
		return -1;
	}

	memset(buffer, 0, sizeof(buffer));

	if ((len=timed_read(fd,buffer,sizeof(buffer)-1,ave_timeout)) == -1) {
		syslog(LOG_ERR, "%s: can't read from AVE socket: %m", id);
		free_array((void **) virus_names,
				sizeof(virus_names) / sizeof(char *));
		ave_disconnect(fd);
		return -1;
	}

	/* it doesn't matter what AVE said at exit so we don't analyze it */

	if (infected == 1) {
		len = 0;
		virus_name = virus_names;
		while (*virus_name) {
			len += strlen(*virus_name) + 1;
			virus_name++;
		}
		len++;
		if ((*viruslist = (char *) malloc(len))) {
			memset(*viruslist, 0, len);
			virus_name = virus_names;
			while (*virus_name) {
				strcat(*viruslist, *virus_name);
				strcat(*viruslist, " ");
				virus_name++;
			}
		}
	}

	free_array((void **) virus_names,
			sizeof(virus_names) / sizeof(char *));

	ave_disconnect(fd);

	return infected;
}

/* this procedure connects to avp control socket and returns
 * it's filedescriptor */

int ave_connect(char * path) {

	int fd = -1;	/* temporary filedescriptor */
	struct sockaddr_un s;

	if (!path || *path == '\0')
		return -1;

	memset(&s, 0 , sizeof(s));	/* clean struct */

	if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
		return -1;

	s.sun_family = AF_UNIX;
	strncpy(s.sun_path, path, sizeof(s.sun_path) - 1);

	if (connect(fd, (struct sockaddr *) &s,
			       sizeof(struct sockaddr_un)) == -1) {
		/* unable to connect to aveserver */
		close(fd);
		fd = -1;
	}

	return fd;
}

/* this procedure closes connection to avp control socket */

void ave_disconnect(int fd) {

	if (fd != -1)
		close(fd);
	return;
}

/* this procedure performs timed read(2) */

ssize_t timed_read(int fd, void * buf, size_t nbytes, int timeout)
{
	fd_set rdfs;
	struct timeval tv;

	tv.tv_sec = timeout;
	tv.tv_usec = 0;

	FD_ZERO(&rdfs);
	FD_SET(fd, &rdfs);
	
	if (select(fd + 1, &rdfs, NULL, NULL, timeout ? &tv : NULL) > 0)
		if (FD_ISSET(fd, &rdfs))
			return read(fd, buf, nbytes);

	errno = ETIMEDOUT;
	return -1;
}

/* this procedure performs timed write(2) */

ssize_t timed_write(int fd, const void * buf, size_t nbytes, int timeout)
{
	fd_set wfds;
	struct timeval tv;

	tv.tv_sec = timeout;
	tv.tv_usec = 0;

	FD_ZERO(&wfds);
	FD_SET(fd, &wfds);

	if (select(fd + 1, NULL, &wfds, NULL, timeout ? &tv : NULL) > 0)
		if (FD_ISSET(fd, &wfds))
			return write(fd, buf, nbytes);

	errno = ETIMEDOUT;
	return -1;
}

/* this routine frees memory occupied by "char * array[xx]" */

void free_array(void ** argv, int argc)
{
	for (; argc; argc--, argv++)
		if (*argv) free (*argv);
}

/* this routine gets string returned by AVE and returns virus name */

char * extract_virus_name(char * line)
{
	char * p;
	int len;
	char * name = NULL;

	line += 4; /* now line points to virus name */
	if ((p = strchr(line, ' '))) {
		len = p - line;
		if ((name = (char *) malloc(len + 1))) {
			memset(name, 0, len + 1);
			strncat(name, line, len);
		}
	}
	return name;
}

/* this routine returns pointer to line from buffer and advances buffer
   pointer up to next line */

char * getline (char ** line)
{
	char * p;

	if (!line || !*line) return NULL;
	while (**line && (**line == '\r' || **line == '\n')) (*line)++;
	p = *line;
	while (**line && **line != '\r' && **line != '\n') (*line)++;
	if (**line) {
		**line = 0;
		(*line)++;
	}
	while (**line && (**line == '\r' || **line == '\n')) (*line)++;
	return *p ? p : NULL ;
}
